import unittest

from orangewidget.tests.base import WidgetTest

from Orange.data import Table, Domain, StringVariable, ContinuousVariable

from orangecontrib.bioinformatics.widgets.OWHomologs import HOMOLOG_ID, OWHomologs
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation


class TestOWMHomologs(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWHomologs)
        self.expected_results = ['12504', '12505', '12506', '16423', '17221']
        id = StringVariable("Entrez ID")
        name = StringVariable("Name")
        attributes_rows = {
            TableAnnotation.gene_as_attr_name: False,
            TableAnnotation.tax_id: "9606",
            TableAnnotation.gene_id_column: "Entrez ID",
        }
        attributes_columns = {
            TableAnnotation.gene_as_attr_name: True,
            TableAnnotation.tax_id: "9606",
            TableAnnotation.gene_id_attribute: "Entrez ID",
        }

        domain = Domain([], (), [name, id])

        data_row = [
            ["CD4", "920"],
            ["CD44", "960"],
            ["CD48", "962"],
            ["CD47", "961"],
            ["CD46", "4179"],
        ]
        self.genes_rows = Table.from_list(domain, data_row)
        self.genes_rows.attributes = attributes_rows

        domain_columns = Domain([ContinuousVariable(name) for name, id in data_row])
        for col, id in zip(domain_columns.attributes, data_row):
            col.attributes["Entrez ID"] = id[1]

        self.genes_columns = Table.from_list(domain_columns, [[1, 2, 1, 2, 1]])
        self.genes_columns.attributes = attributes_columns

    def test_homologs_by_rows(self):

        self.widget.auto_commit = False
        self.widget.target_organism_change(self.widget.taxonomy_ids.index("10090"))

        self.send_signal(self.widget.Inputs.data, self.genes_rows)
        self.widget.auto_commit = True
        self.widget.commit()

        out_data = self.get_output("Genes", self.widget)
        self.assertEqual(len(self.genes_rows), len(out_data))
        mouse_ids = list(out_data.get_column_view(HOMOLOG_ID)[0])
        self.assertListEqual(mouse_ids, self.expected_results)

    def test_auto_commit(self):
        self.send_signal(self.widget.Inputs.data, self.genes_rows)
        out_data = self.get_output("Genes", self.widget)
        self.assertEqual(out_data, None)

    def test_homologs_by_column(self):
        self.widget.auto_commit = False
        self.widget.target_organism_change(self.widget.taxonomy_ids.index("10090"))

        self.send_signal(self.widget.Inputs.data, self.genes_columns)
        self.widget.auto_commit = True
        self.widget.commit()

        out_data = self.get_output("Genes", self.widget)
        mouse_ids = [
            (list(att.attributes.values())[2]) for att in out_data.domain.attributes
        ]

        self.assertEqual(len(self.genes_rows), len(mouse_ids))

        self.assertListEqual(mouse_ids, self.expected_results)


if __name__ == '__main__':
    unittest.main()



from AnyQt.QtCore import QModelIndex
from AnyQt.QtWidgets import QStyleOptionViewItem

from Orange.widgets.gui import LinkRole, LinkStyledItemDelegate


class LinkStyledItemDelegate(LinkStyledItemDelegate):
    def initStyleOption(self, option: QStyleOptionViewItem, index: QModelIndex) -> None:
        super().initStyleOption(option, index)
        if index.data(LinkRole) is not None:
            option.font.setUnderline(True)

    def paint(self, painter, option, index):
        self.initStyleOption(option, index)
        super().paint(painter, option, index)



#!/usr/bin/env python
import setuptools

if __name__ == '__main__':
    setuptools.setup(
        use_scm_version=True,
        setup_requires=['setuptools-scm', 'setuptools>=40.0'],
        install_requires=[
            'Orange3>=3.31.0',
            'orange-widget-base>=4.14.1',
            'scipy>=1.5.0',
            'pyclipper>=1.2.0',
            'point-annotator~=2.0',
            'requests',
            'requests-cache>=0.8.0',
            'serverfiles',
            'resdk>=13.8.0',
            'genesis-pyapi',
            'single_sample_gsea>=0.2.0',
            # Versions are determined by Orange
            'numpy',
        ],
        extras_require={
            # docutils changed html in 0.17; fixing to 0.16 until parser fixed
            # todo: remove docutils when parser fixed in widget-base and released
            'doc': ['sphinx', 'recommonmark', 'sphinx_rtd_theme', 'docutils<0.17'],
            'test': [
                'flake8',
                'flake8-comprehensions',
                'flake8-black',
                'pep8-naming',
                'isort',
                'pre-commit',
                'pytest',
                'coverage',
                'codecov',
            ],
        },
    )



from enum import IntEnum
from datetime import datetime

from dateutil.relativedelta import relativedelta

from AnyQt.QtCore import (
    Qt,
    QSize,
    QObject,
    QAbstractAnimation,
    QPropertyAnimation,
    QParallelAnimationGroup,
    pyqtSignal,
)
from AnyQt.QtWidgets import (
    QFrame,
    QLabel,
    QDialog,
    QComboBox,
    QLineEdit,
    QHBoxLayout,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
)

from Orange.widgets import gui, settings
from Orange.widgets.widget import OWComponent
from Orange.widgets.credentials import CredentialManager

from orangecontrib.bioinformatics.resolwe import (
    GENESIS_PLATFORM,
    RESOLWE_PLATFORM,
    ResolweAuthError,
    genapi,
    resapi,
    connect,
)


def get_credential_manager(server_type: str) -> CredentialManager:
    if server_type == RESOLWE_PLATFORM:
        service_name = resapi.CREDENTIAL_MANAGER_SERVICE
    elif server_type == GENESIS_PLATFORM:
        service_name = genapi.CREDENTIAL_MANAGER_SERVICE
    else:
        raise ValueError('Unexpected server type. Available options: resolwe or genesis.')

    return CredentialManager(service_name)


class SignIn(QDialog):
    def __init__(self, flags, *args, server_type: str = 'resolwe', **kwargs):
        super().__init__(flags, *args, **kwargs)

        self.server_type: str = server_type
        self.cm: CredentialManager = get_credential_manager(server_type)

        self.setWindowTitle('Sign in')
        self.setFixedSize(400, 250)

        self.server_cb_label = QLabel('Server *')
        self.server_cb = QComboBox(self)
        self.server_cb.addItems(resapi.RESOLWE_URLS if server_type == RESOLWE_PLATFORM else [genapi.DEFAULT_URL])
        self.server_cb.setEditable(True)

        self.username_label = QLabel('Username *')
        self.username_line_edit = QLineEdit(self)
        self.username_line_edit.setPlaceholderText('Enter correct username')
        self.username_line_edit.returnPressed.connect(self.sign_in)
        self.username_line_edit.textChanged.connect(self.handle_sign_in_btn)

        self.password_label = QLabel('Password *')
        self.password_line_edit = QLineEdit(self)
        self.password_line_edit.setPlaceholderText('Enter correct password')
        self.password_line_edit.returnPressed.connect(self.sign_in)
        self.password_line_edit.textChanged.connect(self.handle_sign_in_btn)
        self.password_line_edit.setEchoMode(QLineEdit.Password)

        self.sign_in_btn = QPushButton('Sign in', self)
        self.sign_in_btn.setDisabled(True)
        self.sign_in_btn.clicked.connect(self.sign_in)

        self.error_msg = QLabel('Unable to log in with provided credentials.')
        self.error_msg.setStyleSheet('color:red')
        self.error_msg.hide()

        layout = QVBoxLayout(self)
        layout.addWidget(self.server_cb_label)
        layout.addWidget(self.server_cb)
        layout.addWidget(self.username_label)
        layout.addWidget(self.username_line_edit)
        layout.addWidget(self.password_label)
        layout.addWidget(self.password_line_edit)
        layout.addWidget(self.error_msg)
        layout.addStretch()
        layout.addWidget(self.sign_in_btn)

        self.resolwe_instance = None

    def handle_sign_in_btn(self):
        self.sign_in_btn.setEnabled(
            True if self.username_line_edit.text() and self.password_line_edit.text() else False
        )

    def sign_in(self):
        self.server_cb_label.setStyleSheet(None)
        self.username_label.setStyleSheet(None)
        self.password_label.setStyleSheet(None)
        self.error_msg.hide()

        server = self.server_cb.currentText()
        username = self.cm.username if self.cm.username else self.username_line_edit.text()
        password = self.cm.password if self.cm.password else self.password_line_edit.text()

        if not server:
            self.server_cb_label.setStyleSheet('color:red')
            return

        if not username:
            self.username_label.setStyleSheet('color:red')
            return

        if not password:
            self.password_label.setStyleSheet('color:red')
            return

        try:
            self.resolwe_instance = connect(username, password, url=server, server_type=self.server_type)
        except ResolweAuthError:
            self.error_msg.show()
            return

        self.cm.username = username
        self.cm.password = password
        self.accept()


PAST_HOUR = relativedelta(hours=-1)
PAST_DAY = relativedelta(days=-1)
PAST_WEEK = relativedelta(weeks=-1)
PAST_MONTH = relativedelta(months=-1)


class FilterByDateModified(IntEnum):
    any_time = 0
    past_hour = 1
    past_day = 2
    past_week = 3
    past_month = 4

    @staticmethod
    def values():
        now = datetime.now()
        return [None, now + PAST_HOUR, now + PAST_DAY, now + PAST_WEEK, now + PAST_MONTH]

    @staticmethod
    def labels():
        return ['Any time', 'Past hour', 'Past 24 hours', 'Past week', 'Past month']


class SortBy(IntEnum):
    relevance = 0
    name_a_to_z = 1
    name_z_to_a = 2
    contributor_a_to_z = 3
    contributor_z_to_a = 4
    newest_first = 5
    oldest_first = 6

    @staticmethod
    def values():
        return ['', 'name', '-name', 'contributor__last_name', '-contributor__last_name', '-modified', 'modified']

    @staticmethod
    def labels():
        return [
            'Relevance',
            'Name A-Z',
            'Name Z-A',
            'Contributor A-Z',
            'Contributor Z-A',
            'Date modified (newest first)',
            'Date modified (oldest first)',
        ]


class ItemsPerPage(IntEnum):
    min = 0
    med = 1
    max = 2

    @staticmethod
    def values():
        return [20, 50, 100]


class QuantileTransformDist(IntEnum):
    normal = 0
    uniform = 1

    @staticmethod
    def values():
        return ['normal', 'uniform']


class CollapsibleFilterComponent(OWComponent, QObject):
    options_changed = pyqtSignal()

    filter_by_full_text: str = settings.Setting('', schema_only=True)
    filter_by_name: str = settings.Setting('', schema_only=True)
    filter_by_contrib: str = settings.Setting('', schema_only=True)
    filter_by_owner: str = settings.Setting('', schema_only=True)
    filter_by_modified: int = settings.Setting(FilterByDateModified.any_time, schema_only=True)
    sort_by: int = settings.Setting(SortBy.newest_first, schema_only=True)

    FILTER_FULL_TEXT_LABEL = 'Search'
    TOGGLE_BTN_LABEL = 'Narrow your search'
    FILTER_NAME_LABEL = 'Filter by name'
    FILTER_CONTRIB_LABEL = 'Filter by contributor'
    FILTER_OWNER_LABEL = 'Filter by owner'
    FILTER_MODIFIED_LABEL = 'Filter by date modified'
    SORTING_LABEL = 'Sorting'

    def __init__(self, parent_widget, parent_component):
        QObject.__init__(self)
        OWComponent.__init__(self, widget=parent_widget)

        box = gui.widgetBox(parent_component, margin=0)

        self.filter_full_text = gui.lineEdit(
            box,
            self,
            'filter_by_full_text',
            label=self.FILTER_FULL_TEXT_LABEL,
            callback=self.on_filter_full_text_changed,
        )

        self.toggle_animation = QParallelAnimationGroup()
        self.toggle_button = QToolButton()
        self.toggle_button.setCheckable(True)
        self.toggle_button.setChecked(False)
        self.toggle_button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.toggle_button.setArrowType(Qt.RightArrow)
        self.toggle_button.setText(self.TOGGLE_BTN_LABEL)
        self.toggle_button.setStyleSheet('QToolButton {border: none; padding-top: 5px; }')
        self.toggle_button.setIconSize(QSize(15, 15))
        self.toggle_button.pressed.connect(self.on_toggle)

        self.collapsible_components = QScrollArea()
        self.collapsible_components.setMaximumHeight(0)
        self.collapsible_components.setMinimumHeight(0)
        self.collapsible_components.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.collapsible_components.setFrameShape(QFrame.NoFrame)

        box = gui.widgetBox(parent_component, margin=0)
        box.layout().addWidget(self.toggle_button)
        box.layout().addWidget(self.collapsible_components)

        self.toggle_animation.addAnimation(QPropertyAnimation(box, b"minimumHeight"))
        self.toggle_animation.addAnimation(QPropertyAnimation(box, b"maximumHeight"))
        self.toggle_animation.addAnimation(QPropertyAnimation(self.collapsible_components, b"maximumHeight"))

        layout = QHBoxLayout()
        left_box = gui.widgetBox(None, self, margin=0, flat=True)
        mid_box = gui.widgetBox(None, self, margin=0, flat=True)
        right_box = gui.widgetBox(None, self, margin=0, flat=True)

        self.filter_name = gui.lineEdit(
            left_box, self, 'filter_by_name', label=self.FILTER_NAME_LABEL, callback=self.on_filter_changed
        )
        self.filter_contrib = gui.lineEdit(
            mid_box,
            self,
            'filter_by_contrib',
            label=self.FILTER_CONTRIB_LABEL,
            callback=self.on_filter_changed,
        )
        self.filter_owner = gui.lineEdit(
            right_box,
            self,
            'filter_by_owner',
            label=self.FILTER_OWNER_LABEL,
            callback=self.on_filter_changed,
        )
        self.filter_modified = gui.comboBox(
            left_box,
            self,
            'filter_by_modified',
            label=self.FILTER_MODIFIED_LABEL,
            callback=self.on_filter_changed,
            items=FilterByDateModified.labels(),
        )
        self.sorting = gui.comboBox(
            mid_box, self, 'sort_by', label=self.SORTING_LABEL, callback=self.on_filter_changed, items=SortBy.labels()
        )

        gui.rubber(left_box)
        gui.rubber(mid_box)
        gui.rubber(right_box)
        layout.addWidget(left_box)
        layout.addWidget(mid_box)
        layout.addWidget(right_box)
        self.collapsible_components.setLayout(layout)

        collapsed_height = box.layout().sizeHint().height() - self.collapsible_components.maximumHeight()
        content_height = layout.sizeHint().height()

        for i in range(self.toggle_animation.animationCount()):
            animation = self.toggle_animation.animationAt(i)
            animation.setDuration(100)
            animation.setStartValue(collapsed_height)
            animation.setEndValue(collapsed_height + content_height)

        content_animation = self.toggle_animation.animationAt(self.toggle_animation.animationCount() - 1)
        content_animation.setDuration(100)
        content_animation.setStartValue(0)
        content_animation.setEndValue(content_height)

    def on_toggle(self):
        """Start animation"""
        checked = self.toggle_button.isChecked()
        self.toggle_button.setArrowType(Qt.DownArrow if not checked else Qt.RightArrow)
        self.toggle_animation.setDirection(QAbstractAnimation.Forward if not checked else QAbstractAnimation.Backward)
        self.toggle_animation.start()

    def on_filter_full_text_changed(self):
        self.sort_by = SortBy.relevance if self.filter_full_text else SortBy.newest_first
        self.on_filter_changed()

    def on_filter_changed(self):
        self.options_changed.emit()


class PaginationComponent(OWComponent, QObject):
    options_changed = pyqtSignal()

    items_per_page: int = settings.Setting(ItemsPerPage.min, schema_only=True)
    current_page: int = settings.Setting(1, schema_only=True)
    offset: int = settings.Setting(0, schema_only=True)

    PAGE_LIMIT_LABEL = 'Items per page  '

    def __init__(self, parent_widget, parent_component):
        QObject.__init__(self)
        OWComponent.__init__(self, widget=parent_widget)

        box = gui.widgetBox(parent_component, orientation=Qt.Horizontal)

        self.page_limit = gui.radioButtons(
            box,
            self,
            'items_per_page',
            [str(val) for val in ItemsPerPage.values()],
            orientation=Qt.Horizontal,
            callback=self.on_limit_changed,
            label=self.PAGE_LIMIT_LABEL,
        )

        self.page_left_btn = QToolButton()
        self.page_left_btn.setStyleSheet('QToolButton {border: none;}')
        self.page_left_btn.setArrowType(Qt.LeftArrow)
        self.page_left_btn.pressed.connect(self.left_btn_pressed)

        self.offset_label = gui.label(None, self, str(self.current_page), labelWidth=15)
        self.offset_label.setAlignment(Qt.AlignCenter)

        self.page_right_btn = QToolButton()
        self.page_right_btn.setStyleSheet('QToolButton {border: none;}')
        self.page_right_btn.setArrowType(Qt.RightArrow)
        self.page_right_btn.pressed.connect(self.right_btn_pressed)

        parent_widget.pagination_availability.connect(self._handle_paginate_buttons)

        box.layout().addStretch(1)
        box.layout().addWidget(self.page_left_btn)
        box.layout().addWidget(self.offset_label)
        box.layout().addWidget(self.page_right_btn)

    def _handle_paginate_buttons(self, next_page: bool, previous_page: bool):
        self.page_left_btn.setEnabled(previous_page)
        self.page_right_btn.setEnabled(next_page)

    def reset_pagination(self):
        self.offset = 0
        self.current_page = 1
        self.offset_label.setText(str(self.current_page))

    def left_btn_pressed(self):
        new_offset = self.offset - ItemsPerPage.values()[self.items_per_page]

        if new_offset >= 0:
            self.current_page -= 1
            self.offset = new_offset
            self.offset_label.setText(str(self.current_page))

        self.options_changed.emit()

    def right_btn_pressed(self):
        self.current_page += 1
        self.offset += ItemsPerPage.values()[self.items_per_page]
        self.offset_label.setText(str(self.current_page))
        self.options_changed.emit()

    def on_limit_changed(self):
        self.reset_pagination()
        self.options_changed.emit()


class NormalizationComponent(OWComponent, QObject):
    """Gene expression normalization component"""

    options_changed = pyqtSignal()

    quantile_norm: bool
    quantile_norm = settings.Setting(False, schema_only=True)

    log_norm: bool
    log_norm = settings.Setting(True, schema_only=True)

    z_score_norm: bool
    z_score_norm = settings.Setting(False, schema_only=True)

    z_score_axis: int
    z_score_axis = settings.Setting(0, schema_only=True)

    quantile_transform: bool
    quantile_transform = settings.Setting(False, schema_only=True)

    quantile_transform_axis: int
    quantile_transform_axis = settings.Setting(0, schema_only=True)

    quantile_transform_dist: int
    quantile_transform_dist = settings.Setting(0, schema_only=True)

    BOX_TITLE = 'Normalization'
    QUANTILE_NORM_LABEL = 'Quantile normalization'
    LOG_NORM_LABEL = 'Log2(x+1)'
    Z_SCORE_LABEL = 'Z-score'
    QUANTILE_TRANSFORM_LABEL = 'Quantile transform'

    def __init__(self, parent_widget, parent_component):
        QObject.__init__(self)
        OWComponent.__init__(self, widget=parent_widget)

        box = gui.widgetBox(parent_component, self.BOX_TITLE, margin=3)
        gui.checkBox(box, self, 'quantile_norm', self.QUANTILE_NORM_LABEL, callback=self.options_changed.emit)
        gui.checkBox(box, self, 'log_norm', self.LOG_NORM_LABEL, callback=self.options_changed.emit)
        gui.checkBox(box, self, 'z_score_norm', self.Z_SCORE_LABEL, callback=self.on_z_score_selected)
        self.z_score_axis_btn = gui.radioButtons(
            gui.indentedBox(box),
            self,
            'z_score_axis',
            btnLabels=['columns', 'rows'],
            callback=self.options_changed.emit,
            orientation=Qt.Horizontal,
        )
        self.z_score_axis_btn.setHidden(not bool(self.z_score_norm))

        gui.checkBox(
            box,
            self,
            'quantile_transform',
            self.QUANTILE_TRANSFORM_LABEL,
            callback=self.on_quantile_transform_selected,
        )
        self.quantile_transform_axis_btn = gui.radioButtons(
            gui.indentedBox(box),
            self,
            'quantile_transform_axis',
            btnLabels=['columns', 'rows'],
            callback=self.options_changed.emit,
            orientation=Qt.Horizontal,
        )
        self.quantile_transform_dist_btn = gui.radioButtons(
            gui.indentedBox(box),
            self,
            'quantile_transform_dist',
            btnLabels=QuantileTransformDist.values(),
            callback=self.options_changed.emit,
            orientation=Qt.Horizontal,
        )

        self.quantile_transform_axis_btn.setHidden(not bool(self.quantile_transform))
        self.quantile_transform_dist_btn.setHidden(not bool(self.quantile_transform))

    def on_z_score_selected(self):
        self.z_score_axis_btn.setHidden(not bool(self.z_score_norm))
        self.options_changed.emit()

    def on_quantile_transform_selected(self):
        self.quantile_transform_axis_btn.setHidden(not bool(self.quantile_transform))
        self.quantile_transform_dist_btn.setHidden(not bool(self.quantile_transform))
        self.options_changed.emit()



from orangecontrib.bioinformatics import go

ontology = go.Ontology()

# Load annotations for yeast.
annotations = go.Annotations("4932", ontology=ontology)

# keys are symbol names, values are Entrez IDs
genes = {'RRB1': '855161', 'OST4': '851366', 'VID27': '855509'}
res = annotations.get_enriched_terms(genes.values())


print(annotations.gene_annotations['855161'])
for a in annotations.gene_annotations['855161']:
    print(ontology[a.go_id].name + " with evidence code " + a.evidence)


# Get all genes annotated to the same terms as RRB1
ids = set([a.go_id for a in annotations.gene_annotations['855161']])
for term_id in ids:
    ants = annotations.get_annotations_by_go_id(term_id)
    genes = set([a.gene_id for a in ants])
    print(", ".join(genes) + " annotated to " + term_id + " " + ontology[term_id].name)




"""
==============================================
KEGG - Kyoto Encyclopedia of Genes and Genomes
==============================================

:mod:`kegg` is a python module for accessing `KEGG (Kyoto Encyclopedia
of Genes and Genomes) <http://www.genome.jp/kegg/>`_ using its web services.

.. note:: This module requires `slumber`_ and `requests`_ packages.

.. _`slumber`: https://pypi.python.org/pypi/slumber/

.. _`requests`: https://pypi.python.org/pypi/requests

>>> # Create a KEGG Genes database interface
>>> genome = KEGGGenome()
>>> # List all available entry ids
>>> keys = list(genome.keys())
>>> print(keys[0])
T01001
>>> # Retrieve the entry for the key.
>>> entry = genome[keys[0]]
>>> print(entry.entry_key)
T01001
>>> print(entry.definition)
Homo sapiens (human)
>>> print(entry)  # doctest: +SKIP
ENTRY       T01001            Complete  Genome
NAME        hsa, HUMAN, 9606
DEFINITION  Homo sapiens (human)
...

The :class:`Organism` class can be a convenient starting point
for organism specific databases.

>>> organism = Organism("Homo sapiens")  # searches for the organism by name
>>> print(organism.org_code)  # prints the KEGG organism code
hsa
>>> genes = organism.genes  # get the genes database for the organism
>>> gene_ids = list(genes.keys()) # KEGG gene identifiers
>>> entry = genes["hsa:672"]
>>> print(entry.definition) # doctest: +SKIP
(RefSeq) BRCA1, DNA repair associated
>>> # print the entry in DBGET database format.
>>> print(entry) # doctest: +SKIP
ENTRY       672               CDS       T01001
NAME        BRCA1, BRCAI, BRCC1, BROVCA1, FANCS, IRIS, PNCA4, PPP1R53, PSCP, RNF53
DEFINITION  ...
"""
from __future__ import absolute_import

import os
import sys
import threading
from datetime import datetime
from functools import reduce
from itertools import chain
from contextlib import contextmanager
from collections import defaultdict

from orangecontrib.bioinformatics.kegg import api, conf, entry, pathway, databases
from orangecontrib.bioinformatics.ncbi import taxonomy
from orangecontrib.bioinformatics.utils import statistics, progress_bar_milestones
from orangecontrib.bioinformatics.kegg.brite import Brite, BriteEntry

KEGGGenome = databases.Genome
KEGGGenes = databases.Genes
KEGGEnzyme = databases.Enzyme
KEGGReaction = databases.Reaction
KEGGPathways = databases.Pathway
KEGGCompound = databases.Compound

KEGGBrite = Brite
KEGGBriteEntry = BriteEntry

KEGGPathway = pathway.Pathway

DEFAULT_CACHE_DIR = conf.params["cache.path"]


class Organism(object):
    """
    A convenience class for retrieving information regarding an
    organism in the KEGG Genes database.

    :param str org: KEGG organism code (e.g. "hsa", "sce"). Can also be a
        descriptive name (e.g. 'yeast', "homo sapiens") in which case the
        organism code will be searched for by using KEGG `find` api.

    .. seealso::

        :func:`organism_name_search`
            Search KEGG for an organism code

    """

    def __init__(self, org):
        self.org_code = self.organism_name_search(org)
        self.api = api.CachedKeggApi()

    @property
    def org(self):
        """
        KEGG organism code.
        """
        return self.org_code

    @property
    def genes(self):
        """
        An :class:`~.databases.Genes` database instance for this organism.
        """
        # TODO: This should not be a property but a method.
        # I think it was only put here as back compatibility with old obiKEGG.
        if not hasattr(self, "_genes"):
            genes = KEGGGenes(self.org_code)
            self._genes = genes
        return self._genes

    def kegg_to_ncbi_mapper(self):
        return {kegg.upper(): ncbi.split(':')[-1] for ncbi, kegg in self.api.conv(self.org_code, "ncbi-geneid")}

    def get_ncbi_ids(self):
        # return [ncbi.split(':')[-1] for ncbi, _ in self.api.conv(self.org_code, "ncbi-geneid")]
        return [ncbi for ncbi in self.kegg_to_ncbi_mapper().values()]

    def gene_aliases(self):
        """
        Return a list of sets of equal genes (synonyms) in KEGG for
        this organism.

        .. note::

            This only includes 'ncbi-geneid' and 'ncbi-proteinid' records
            from the KEGG Genes DBLINKS entries.

        """
        definitions = self.api.list(self.org_code)
        ncbi_geneid = self.api.conv(self.org_code, "ncbi-geneid")
        ncbi_gi = self.api.conv(self.org_code, "ncbi-proteinid")

        aliases = defaultdict(set)

        for entry_id, definition in definitions:
            # genes entry id without the organism code
            aliases[entry_id].add(entry_id.split(":", 1)[1])
            # all names in the NAME field (KEGG API list returns
            # 'NAME; DEFINITION') fields for genes
            names = definition.split(";")[0].split(",")
            aliases[entry_id].update([name.strip() for name in names])

        for source_id, target_id in chain(ncbi_geneid, ncbi_gi):
            aliases[target_id].add(source_id.split(":", 1)[1])

        return [set([entry_id]).union(names) for entry_id, names in aliases.items()]

    def pathways(self, with_ids=None):
        """
        Return a list of all pathways for this organism.
        """
        if with_ids is not None:
            return self.api.get_pathways_by_genes(with_ids)
        else:
            return [p.entry_id for p in self.api.list_pathways(self.org_code)]

    def list_pathways(self):
        """
        List all pathways for this organism.

        .. deprecated: 2.5
            Use :func:`pathways` instead.

        """
        # NOTE: remove/deprecate and use pathways()
        return self.pathways()

    def get_linked_pathways(self, pathway_id):
        self.api.get_linked_pathways(pathway_id)

    def enzymes(self, genes=None):
        raise NotImplementedError()

    def get_enriched_pathways(self, genes, reference=None, prob=statistics.Binomial(), callback=None):
        """
        Return a dictionary with enriched pathways ids as keys
        and (list_of_genes, p_value, num_of_reference_genes) tuples
        as items.

        """
        if reference is None:
            reference = self.genes.keys()
        reference = set(reference)

        allPathways = defaultdict(lambda: [[], 1.0, []])
        milestones = progress_bar_milestones(len(genes), 100)
        pathways_db = KEGGPathways()

        pathways_for_gene = []
        for i, gene in enumerate(genes):
            pathways_for_gene.append(self.pathways([gene]))
            if callback and i in milestones:
                callback(i * 50.0 / len(genes))

        # pre-cache for speed
        pathways_db.pre_cache([pid for pfg in pathways_for_gene for pid in pfg])
        for i, (gene, pathways) in enumerate(zip(genes, pathways_for_gene)):
            for pathway in pathways:
                if pathways_db.get_entry(pathway).gene:
                    allPathways[pathway][0].append(gene)
            if callback and i in milestones:
                callback(50.0 + i * 50.0 / len(genes))

        pItems = allPathways.items()

        for i, (p_id, entry) in enumerate(pItems):
            pathway = pathways_db.get_entry(p_id)
            entry[2].extend(reference.intersection(pathway.gene or []))
            entry[1] = prob.p_value(len(entry[0]), len(reference), len(entry[2]), len(genes))
        return dict([(pid, (genes, p, len(ref))) for pid, (genes, p, ref) in allPathways.items()])

    def get_genes_by_enzyme(self, enzyme):
        enzyme = KEGGEnzyme().get_entry(enzyme)
        return enzyme.genes.get(self.org_code, []) if enzyme.genes else []

    def get_genes_by_pathway(self, pathway_id):
        # print(len(KEGGPathway(pathway_id).genes()), KEGGPathway(pathway_id).genes()[1])
        return KEGGPathway(pathway_id).genes()

    def get_enzymes_by_pathway(self, pathway_id):
        return KEGGPathway(pathway_id).enzymes()

    def get_compounds_by_pathway(self, pathway_id):
        return KEGGPathway(pathway_id).compounds()

    def get_pathways_by_genes(self, gene_ids):
        """ Pathways that include all genes in gene_ids. """
        l = self.api.get_genes_pathway_organism(self.org_code)
        gene_ids = set(gene_ids)
        gtp = defaultdict(set)
        for a, b in l:
            gtp[a].add(b)
        pathways = [gtp[g] for g in gene_ids]
        pathways = reduce(set.intersection, pathways)
        return sorted(pathways)

    def get_pathways_by_enzymes(self, enzyme_ids):
        enzyme_ids = set(enzyme_ids)
        pathways = [KEGGEnzyme()[id].pathway for id in enzyme_ids]
        pathways = reduce(set.union, pathways, set())
        return [id for id in pathways if enzyme_ids.issubset(KEGGPathway(id).enzymes())]

    def get_pathways_by_compounds(self, compound_ids):
        compound_ids = set(compound_ids)
        pathways = [KEGGCompound()[id].pathway for id in compound_ids]
        pathways = reduce(set.union, pathways, set())
        return [id for id in pathways if compound_ids.issubset(KEGGPathway(id).compounds())]

    def get_enzymes_by_compound(self, compound_id):
        return KEGGCompound()[compound_id].enzyme

    def get_enzymes_by_gene(self, gene_id):
        return self.genes[gene_id].enzymes

    def get_compounds_by_enzyme(self, enzyme_id):
        return self._enzymes_to_compounds.get(enzyme_id)

    def get_genes(self):
        return self.genes

    @classmethod
    def organism_name_search(cls, name):
        return organism_name_search(name)

    @classmethod
    def organism_version(cls, name):
        name = cls.organism_name_search(name)
        with _global_genome_instance() as genome:
            info = genome.api.info(name)
            return info.releas


KEGGOrganism = Organism


def organism_name_search(name):
    """
    Search for a organism by `name` and return it's KEGG organism code.
    """
    with _global_genome_instance() as genome:
        try:
            name = genome.org_code_to_entry_key(name)
        except ValueError:
            pass

        if name not in genome:
            ids = genome.search(name)
            if ids:
                name = ids.pop(0) if ids else name
            else:
                raise taxonomy.UnknownSpeciesIdentifier(name)

        try:
            return genome[name].organism_code
        except Exception as e:
            raise taxonomy.UnknownSpeciesIdentifier(name)


def pathways(org):
    """
    Return a list of all KEGG pathways for an KEGG organism code `org`.
    """
    return KEGGPathway.list(org)


def from_taxid(taxid):
    """
    Return a KEGG organism code for a NCBI Taxonomy id string `taxid`.
    """
    with _global_genome_instance() as genome:
        # NOTE should link using https://www.genome.jp/linkdb/linkdb_api.html
        # e.g http://rest.genome.jp/link/tax/gn
        taxid = genome.TAXID_MAP.get(taxid, taxid)
        for code in genome.essential_organisms() + genome.common_organisms():
            entry_ = genome.get(code, None)
            if entry_ is not None and entry_.taxid == taxid:
                return entry_.org_code

        res = genome.search(taxid)
        for r in res:
            e = genome[r]

            if e.taxid in [taxid, genome.TAXID_MAP.get(taxid, taxid)]:
                return e.org_code()

        return None


def to_taxid(name):
    """
    Return a NCBI Taxonomy id for a given KEGG Organism name
    """
    with _global_genome_instance() as genome:

        if name in genome:  # a T string
            return genome[name].taxid

        name2 = genome.org_code_to_entry_key(name)
        if name2 in genome:
            return genome[name2].taxid

        keys = genome.search(name)
        if keys:
            return genome[keys[0]].taxid
        else:
            return None


def _global_genome_instance():
    genome = getattr(_global_genome_instance, "_genome", None)
    if genome is None:
        genome = KEGGGenome()
        genome._lock = threading.RLock()
        _global_genome_instance._genome = genome

    @contextmanager
    def instance_locked(instance, lock):
        lock.acquire()
        try:
            yield instance
        finally:
            lock.release()

    return instance_locked(genome, genome._lock)



""" NCBI Taxonomy browser module """

from orangecontrib.bioinformatics.ncbi.taxonomy.utils import Taxonomy, UnknownSpeciesIdentifier

COMMON_NAMES = (
    ("6500", "Aplysia californica"),
    ("3702", "Arabidopsis thaliana"),
    ("9913", "Bos taurus"),
    ("6239", "Caenorhabditis elegans"),
    ("5476", "Candida albicans"),
    ("3055", "Chlamydomonas reinhardtii"),
    ("7955", "Danio rerio"),
    ("44689", "Dictyostelium discoideum"),
    ("7227", "Drosophila melanogaster"),
    ("562", "Escherichia coli"),
    ("11103", "Hepatitis C virus"),
    ("9606", "Homo sapiens"),
    ("10090", "Mus musculus"),
    ("2104", "Mycoplasma pneumoniae"),
    ("4530", "Oryza sativa"),
    ("5833", "Plasmodium falciparum"),
    ("4754", "Pneumocystis carinii"),
    ("10116", "Rattus norvegicus"),
    ("4932", "Saccharomyces cerevisiae"),
    ("4896", "Schizosaccharomyces pombe"),
    ("31033", "Takifugu rubripes"),
    ("8355", "Xenopus laevis"),
    ("4577", "Zea mays"),
)

SHORT_NAMES = {
    "6500": ["aplysia"],
    "3702": ["arabidopsis", "thaliana", "plant"],
    "9913": ["cattle", "cow"],
    "6239": ["nematode", "roundworm"],
    "5476": ["thrush", "candidiasis", "candida"],
    "3055": ["algae"],
    "7955": ["zebrafish"],
    "44689": ["dicty", "amoeba", "slime mold"],
    "7227": ["fly", "fruit fly", "vinegar fly"],
    "562": ["ecoli", "coli", "bacterium"],
    "11103": ["virus, hepatitis"],
    "9606": ["human"],
    "10090": ["mouse", "mus"],
    "2104": ["bacterium", "mycoplasma"],
    "4530": ["asian rice", "rice", "cereal", "plant"],
    "5833": ["plasmodium", "malaria", "parasite"],
    "4754": ["pneumonia", "fungus"],
    "10116": ["rat", "laboratory rat"],
    "4932": ["yeast", "baker yeast", "brewer yeast"],
    "4896": ["yeast", "fission yeast"],
    "31033": ["fish", "pufferfish"],
    "8355": ["frog", "african clawed frog"],
    "4577": ["corn", "cereal grain", "plant"],
}


COMMON_NAMES_MAPPING = dict(COMMON_NAMES)


def common_taxids():
    """ Return taxonomy IDs for most common organisms.

    :rtype: :class:`list` of common organisms
    """
    return [tax_id for tax_id, _ in COMMON_NAMES]


def common_taxid_to_name(tax_id):
    """ Return a name for a common organism taxonomy id.

    :param tax_id: Taxonomy if (NCBI taxonomy database)
    :type tax_id: str

    :rtype: :class:`str`
    """
    if str(tax_id) == '352472':
        tax_id = '44689'
    return COMMON_NAMES_MAPPING[str(tax_id)]


def species_name_to_taxid(organism_name):
    """ Return taxonomy ID for a taxonomy name.

    :param organism_name: Official organism name, e.g., 'Homo sapiens'
    :type organism_name: str

    :rtype: :class:`str`
    """
    name_to_taxid = dict(map(reversed, COMMON_NAMES))

    if organism_name in name_to_taxid:
        return name_to_taxid[organism_name]
    return None


def shortname(tax_id):
    """ Short names for common_taxids organisms

    :param tax_id: Taxonomy if (NCBI taxonomy database)
    :type tax_id: str

    """
    if tax_id in SHORT_NAMES:
        return SHORT_NAMES[tax_id]
    return []


def name(tax_id):
    """ Return the scientific name for organism with provided taxonomy id.

    :param tax_id: Taxonomy if (NCBI taxonomy database)
    :type tax_id: str

    """
    # Most of the lookups will be for the common names, so in most
    # situations we can avoid loading the taxonomy.
    if tax_id in COMMON_NAMES_MAPPING:
        return COMMON_NAMES_MAPPING[tax_id]
    else:
        return Taxonomy()[tax_id]


def other_names(tax_id):
    """ Return a list of (name, name_type) tuples excluding the scientific name.

    Use :func:`name` to retrieve the scientific name.

    :param tax_id: Taxonomy if (NCBI taxonomy database)
    :type tax_id: str

    """
    return Taxonomy().other_names(tax_id)


def search(string, only_species=True, exact=False):
    """ Search the NCBI taxonomy database for an organism.

    :param string: Search string.
    :type string: str

    :param only_species: Return only taxids of species (and subspecies).
    :type only_species: bool

    :param exact:  Return only taxids of organism that exactly match the string.
    :type exact: bool
    """
    ids = Taxonomy().search(string, only_species, exact)
    return list(ids)


def lineage(tax_id):
    """ Return a list of taxids ordered from the topmost node (root) to taxid.

    :param tax_id: Taxonomy if (NCBI taxonomy database)
    :type tax_id: str
    """

    return Taxonomy().lineage(tax_id)



""" OWMarkerGenes """
from typing import List, Tuple, Iterable, Optional
from functools import partial

import numpy as np
import requests
from orangewidget.settings import Setting, ContextHandler

from AnyQt import QtGui, QtCore
from AnyQt.QtCore import Qt, QSize, QMimeData, QModelIndex, QAbstractItemModel, QSortFilterProxyModel, pyqtSignal
from AnyQt.QtWidgets import QLayout, QWidget, QLineEdit, QTreeView, QGridLayout, QTextBrowser, QAbstractItemView

from Orange.data import Table, RowInstance
from Orange.widgets import gui, widget, settings

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID, GENE_ID_COLUMN, GENE_AS_ATTRIBUTE_NAME
from orangecontrib.bioinformatics.widgets.utils.settings import MarkerGroupContextHandler

SERVER_FILES_DOMAIN = 'marker_genes'
GROUP_BY_ITEMS = ["Cell Type", "Function"]
FILTER_COLUMNS_DEFAULT = ["Name", "Entrez ID"]
NUM_LINES_TEXT = 5
MAP_GROUP_TO_TAX_ID = {'Human': '9606', 'Mouse': '10090'}


class TreeItem(object):
    def __init__(
        self, name: str, is_type_item: bool, data_row: Optional[RowInstance], parent: Optional["TreeItem"] = None
    ):
        self.name: str = name
        self.parentItem: Optional["TreeItem"] = None
        self.childItems: List = []
        self.change_parent(parent)
        self.isTypeItem: bool = is_type_item

        self.data_row: Optional[RowInstance] = data_row
        # holds the information whether element is expanded in the tree
        self.expanded = False

    def change_parent(self, parent: Optional["TreeItem"] = None) -> None:
        """
        Function replaces the item parent.

        Parameters
        ----------
        parent
            Parent which is set as an items parent
        """
        self.parentItem = parent
        if parent is not None:
            parent.append_child(self)

    @staticmethod
    def change_parents(items: Iterable["TreeItem"], parent: "TreeItem") -> None:
        """
        This function assigns all items a parent

        Parameters
        ----------
        items
            Items that will get parent assigned
        parent
            Parent that will be assigned
        """
        for it in items:
            it.change_parent(parent)

    def append_child(self, item: "TreeItem") -> None:
        """
        This function appends child to self. This function just append a child
        and do not set parent of child item (it is set by function calling this function.

        Parameters
        ----------
        item
            Child item.
        """
        self.childItems.append(item)

    def child(self, row: int) -> "TreeItem":
        """
        Returns row-th child the item.

        Parameters
        ----------
        row
            Index of the child

        Returns
        -------
        Child item
        """
        return self.childItems[row]

    def remove_children(self, position: int, count: int) -> bool:
        """
        This function removes children form position to position + count

        Parameters
        ----------
        position
            Position of the first removed child
        count
            Number of removed children
        Returns
        -------
        Success of removing
        """
        # we could check for every index separately but in case when all items cannot be removed there is something
        # wrong with the cal - just refuse to do the removal
        if position < 0 or position + count > len(self.childItems):
            return False
        # remove from the back otherwise you get index out of range when removing more than one
        for i in range(position + count - 1, position - 1, -1):
            self.childItems.pop(i)
        return True

    def children_count(self) -> int:
        """
        Function return number fo children of an item.

        Returns
        -------
        Number of children.
        """
        return len(self.childItems)

    def row(self) -> int:
        """
        Function returns the index of the item in a parents list. If it does not have parent returns 0.

        Returns
        -------
        The index of an item.
        """
        if self.parentItem:
            return self.parentItem.childItems.index(self)
        return 0

    def child_from_name_index(self, name: str) -> Tuple[Optional[int], Optional["TreeItem"]]:
        """
        This function returns the index (in the array of children) and item with the name specified.

        Parameters
        ----------
        name
            The name of requested item.

        Returns
        -------
        The index and the searched item. If it does not exist return None.
        """
        for i, c in enumerate(self.childItems):
            if c.name == name:
                return i, c
        return None, None

    def contains_text(self, text: str, filter_columns: List[str]) -> bool:
        """
        Function indicates whether the text is present in any of columns in FILTER_COLUMNS.

        Parameters
        ----------
        text
            Tested text.
        filter_columns
            Columns used in filtering

        Returns
        -------
        Boolean that indicates that text is present in the item.
        """
        if self.data_row is not None:
            return any(text.lower() in str(self.data_row[col]).lower() for col in filter_columns)
        else:
            return False

    def get_data_rows(self) -> List[RowInstance]:
        """
        This function returns a list of rows present in self and any children node.

        Returns
        -------
        Rows with data items.
        """
        rows = [self.data_row] if self.data_row is not None else []
        for c in self.childItems:
            rows += c.get_data_rows()
        return rows

    def count_marker_genes(self) -> int:
        """
        This function counts number of marker genes in its tree

        Returns
        -------
        int
            Number of marker genes in the tree including itself.
        """
        return sum(c.count_marker_genes() for c in self.childItems) + (0 if self.data_row is None else 1)


class TreeModel(QAbstractItemModel):

    MIME_TYPE = "application/x-Orange-TreeModelData"
    # it was slow to relay on the qt model signals since they are emitted every time endInsertRows is called
    # this one will be emitted just a the end
    data_added = pyqtSignal()
    data_removed = pyqtSignal()
    expand_item = pyqtSignal(QModelIndex, bool)

    def __init__(self, data: Table, parent_column: str, parent: QModelIndex = None) -> None:
        super().__init__(parent)

        self.rootItem = TreeItem("Genes", False, None)
        self.setup_model_data(data, self.rootItem, parent_column)
        self.filter_columns = FILTER_COLUMNS_DEFAULT + [parent_column]

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDropEnabled
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDragEnabled

    def supportedDropActions(self) -> Qt.DropAction:  # noqa: N802
        return Qt.MoveAction

    def supportedDragActions(self) -> Qt.DropAction:  # noqa: N802
        return Qt.MoveAction

    def columnCount(self, parent: QModelIndex = None) -> int:  # noqa: N802
        """
        Our TreeView has only one column
        """
        return 1

    def rowCount(self, parent: QModelIndex = None) -> int:  # noqa: N802
        """
        Function returns the number children of an item.
        """
        if not parent or not parent.isValid():
            parent_item = self.rootItem
        else:
            parent_item = self.node_from_index(parent)

        count = parent_item.children_count()
        return count

    def data(self, index: QModelIndex, role=None) -> Optional[str]:
        """
        Function returns the index's data that are shown in the view.
        """
        if not index.isValid() or role != Qt.DisplayRole:
            return None
        item = self.node_from_index(index)
        rd = item.name
        return rd

    def headerData(self, section, orientation, role=None):  # noqa: N802
        """
        Return header data (they are no used so function is just here to make Qt happy).
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.rootItem.name
        return None

    def index(self, row: int, column: int, parent: QModelIndex = None) -> QModelIndex:
        """
        Function create QModelIndex from the parent item and row and column index.

        Parameters
        ----------
        row
            Row of the child
        column
            Column of th child - not used since we have only one column.
        parent
            Parent of the child we are making index of.

        Returns
        -------
        Index of the child at row-th row of the parent.
        """
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parent_item = self.rootItem
        else:
            parent_item = self.node_from_index(parent)

        child_item = parent_item.child(row)
        if child_item:
            return self.createIndex(row, column, child_item)
        else:
            return QModelIndex()

    def parent(self, index: QModelIndex) -> QModelIndex:
        """
        Function returns parent of the item in index.
        """
        if not index.isValid():
            return QModelIndex()

        child_item = self.node_from_index(index)
        parent_item = child_item.parentItem if child_item else None

        if not parent_item or parent_item == self.rootItem:
            return QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def removeRows(  # noqa: N802
        self, position: int, rows: int, parent: QModelIndex = QModelIndex(), emit_data_removed: bool = True
    ) -> bool:
        """
        This function implements removing rows form position to position + rows form the model. It also removes
        a parent when it becomes empty. If remove is called over the parent it remove children together with the
        parent.

        Parameters
        ----------
        emit_data_removed
            It indicates whether we emit dataRemoved signal. It makes possible that it is emitted when function called
            by Qt and not when called by removeNodeList. When qt call this function indices are optimized to be in
            range and this function is called just once for consecutive data. It is not true for data from
            removeNodeList. In this case we call removeRows for each removed item and it can become slow when many
            items are removed.
        """
        node = self.node_from_index(parent)
        if not node:
            return False
        if not (position == 0 and node.children_count() == rows) or not node.isTypeItem:
            # when not all children are removed or when type item is removed - when removed node is a type item its
            # parent is not
            self.beginRemoveRows(QModelIndex() if node is self.rootItem else parent, position, position + rows - 1)
            success = node.remove_children(position, rows)
        else:
            # when node is a children of a type item and all children of node needs to be removed, remove node itself
            self.beginRemoveRows(QModelIndex(), node.row(), node.row())
            success = self.rootItem.remove_children(node.row(), 1)
        self.endRemoveRows()
        if emit_data_removed:
            self.data_removed.emit()
        return success

    def mimeTypes(self) -> List[str]:  # noqa: N802
        return [self.MIME_TYPE]

    def mimeData(self, index_list: List[QModelIndex]) -> QMimeData:  # noqa: N802
        """
        This function packs data in QMimeData in order to move them to other view with drag and drop functionality.
        """
        items = [self.node_from_index(index) for index in index_list]
        mime = QMimeData()
        # the encoded 'data' is empty, variables are passed by properties
        mime.setData(self.MIME_TYPE, b'')
        mime.setProperty("_items", items)
        mime.setProperty("_source_id", id(self))
        return mime

    def dropMimeData(  # noqa: N802
        self, mime: QMimeData, action: Qt.DropAction, row: int, column: int, parent: QModelIndex = QModelIndex
    ) -> bool:
        """
        This function unpacks QMimeData and insert them to the view where they are dropped.
        Inserting is a bit complicated since sometimes we need to create a parent when it is not in the view yet:
        - when child (marker gene) is dropped we insert it under the current parent (group) if it exist othervise
          we crate a new parent.
        - when group item is inserted it merge it with current same parent item if exist or insert it if it does
          not exists
        """
        if action == Qt.IgnoreAction:
            return True
        if not mime.hasFormat(self.MIME_TYPE):
            return False
        items = mime.property("_items")
        if items is None:
            return False
        # when parent item in items remove its child since all group will be moved
        items = [it for it in items if it.parentItem not in items]
        self.insert_items(items)
        self.data_added.emit()
        return True

    def insert_items(self, items: List[TreeItem]) -> None:
        """
        This function goes through all items and insert them in the view.
        Parameters
        ----------
        items
            List of items that will be inserted.
        """
        for item in items:
            if item.isTypeItem:
                # inserting group item
                self.insert_group(item)
            else:
                # inserting children (genes)
                self.insert_child_item(item)

    def insert_group(self, item: TreeItem) -> None:
        """
        This function makes insert of the item in the model in case that it is a group item.

        Parameters
        ----------
        item
            Item that is inserted - group item (not gene).
        """
        expand = None

        item_curr_idx, curr_item = self.rootItem.child_from_name_index(item.name)
        if item_curr_idx is None:
            # it is not in the view yet - just insert it
            self.beginInsertRows(QModelIndex(), self.rootItem.children_count(), self.rootItem.children_count())
            item.change_parent(self.rootItem)
            expand = item.expanded
        else:
            # it is in the view already - insert only children
            tr_idx = self.createIndex(item_curr_idx, 0, curr_item)
            self.beginInsertRows(tr_idx, self.rowCount(tr_idx), self.rowCount(tr_idx) + len(item.childItems) - 1)
            TreeItem.change_parents(item.childItems, curr_item)
        self.endInsertRows()

        # emit the signal if item must be expanded
        if expand is not None:
            self.expand_item.emit(self.index_from_node(item), expand)

    def insert_child_item(self, item: TreeItem) -> None:
        """
        This function makes insert of the item in the model in case that it is a child (gene) item.

        Parameters
        ----------
        item
            Item that is inserted - child item (gene).
        """
        expand_parent = None
        parent_curr_idx, curr_parent = self.rootItem.child_from_name_index(item.parentItem.name)
        if parent_curr_idx is None:
            # parent (group) do not exists yet - crate it
            self.beginInsertRows(QModelIndex(), self.rootItem.children_count(), self.rootItem.children_count())
            curr_parent = TreeItem(item.parentItem.name, True, None, self.rootItem)
            # when creating new parent because child is inserted always expend since it is expanded in the other view
            expand_parent = True
        else:
            # parent (group) exists - insert into it
            tr_idx = self.createIndex(parent_curr_idx, 0, curr_parent)
            self.beginInsertRows(tr_idx, self.rowCount(tr_idx), self.rowCount(tr_idx))
        item.change_parent(curr_parent)
        self.endInsertRows()

        # emit the signal if parent must be expanded
        if expand_parent is not None:
            self.expand_item.emit(self.index_from_node(curr_parent), expand_parent)

    def canDropMimeData(  # noqa: N802
        self, data: QMimeData, action: Qt.DropAction, row: int, column: int, parent: QModelIndex
    ) -> bool:
        """
        This function enable or disable drop action to the view. With current implementation we disable drop action to
        the same view.
        """
        if data.property("_source_id") == id(self):
            return False
        else:
            return True

    def remove_node_list(self, indices: List[QModelIndex]) -> None:
        """
        Sometimes we need to remove items provided as a list of QModelIndex. Since it is not possible directly with
        removeRows it is implemented with this function.

        Parameters
        ----------
        indices
            List of nodes as QModelIndex
        """
        for idx in indices:
            node = self.node_from_index(idx)
            if node in node.parentItem.childItems:  # it is possible that item already removed since no children left
                parent_idx = self.index_from_node(node.parentItem)
                self.removeRows(node.row(), 1, parent_idx, emit_data_removed=False)
        self.data_removed.emit()

    def node_from_index(self, index: QModelIndex) -> TreeItem:  # noqa: N802
        """
        Function returns a TreeItem saved in the index.

        Parameters
        ----------
        index
            Index of the TreeItem

        Returns
        -------
            Tree item from the index.
        """
        if index.isValid():
            return index.internalPointer()
        else:
            return self.rootItem

    def index_from_node(self, node: TreeItem) -> QModelIndex:  # noqa: N802
        """
        This function is similar to the index function. We use it in case when we create index but not base on the
        parent index. We could call index but it is faster to create it directly.

        Parameters
        ----------
        node
            TreeItem for which we crate index.

        Returns
        -------
        QModelIndex for TreeItem
        """
        return self.createIndex(node.row(), 0, node)

    def index_from_name(self, name: str, parent: TreeItem) -> Optional[QModelIndex]:  # noqa: N802
        """
        Function create index from tree item name.

        Parameters
        ----------
        name
            Name of the item
        parent
            Parent TreeItem

        Returns
        -------
        QModelIndex for an item
        """
        i, node = parent.child_from_name_index(name)
        if i is None:
            return None
        return self.createIndex(i, 0, node)

    def setup_model_data(self, data_table: Table, parent: TreeItem, parent_column: str) -> None:  # noqa: N802
        """
        Function populates the view with the items from the data table. Items are inserted as a tree with the depth 2
        each row (gene from the table) get inserted under the parent item (group) which is defined with the
        paren_column parameter (e.g. marker gene).

        Parameters
        ----------
        data_table
            Data table with marker genes.
        parent
            Parent under which items are inserted (usually root item of the model).
        parent_column
            Column which is the group for items in the tree.
        """
        parents_dict = {}
        names = data_table.get_column_view("Name")[0]
        types = data_table.get_column_view(parent_column)[0]
        for n, pt, row in zip(names, types, data_table):
            if pt not in parents_dict:
                parents_dict[pt] = TreeItem(pt, True, None, parent)

            TreeItem(n, False, row, parents_dict[pt])

    def set_expanded(self, index: QModelIndex, expanded: bool) -> None:
        """
        Sets the expanded flag of the item. This flag tell whether the item in the view is expanded.

        Parameters
        ----------
        index
            Index that is expanded/collapsed
        expanded
            If true expanded otherwise collapsed
        """
        node = self.node_from_index(index)
        node.expanded = expanded

    def __len__(self):
        """
        Len function in this case returns number of marker genes in the view.

        Returns
        -------
        int
            Number of marker genes in the view.
        """
        return self.rootItem.count_marker_genes()


class FilterProxyModel(QSortFilterProxyModel):
    """
    FilterProxyModel is used to sort items in alphabetical order, and to make filter on the view of available genes.

    Parameters
    ----------
    filter_line_edit
        Line edit used to filter the view.
    """

    def __init__(self, filter_line_edit: QLineEdit) -> None:
        super().__init__()
        self.filter_line_edit = filter_line_edit
        if filter_line_edit is not None:
            filter_line_edit.textChanged.connect(self.setFilterFixedString)

        # have marker genes in the alphabetical order
        self.sort(0)

    def filterAcceptsRow(self, p_int: int, index: QModelIndex) -> bool:  # noqa: N802
        """
        Indicates whether the item should be shown based on the the filter string.
        """
        if self.filter_line_edit is None:
            return True
        model = self.sourceModel()
        idx = model.index(p_int, 0, index)
        res = model.node_from_index(idx).contains_text(self.filter_line_edit.text(), self.sourceModel().filter_columns)

        if model.hasChildren(idx):
            num_items = model.rowCount(idx)
            for i in range(num_items):
                res = res or self.filterAcceptsRow(i, idx)
        return res


class TreeView(QTreeView):
    """
    QTreeView with some additional settings.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setHeaderHidden(True)

        # drag and drop enabled
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setUniformRowHeights(True)

        # the view needs to be aware of the other view for the expanding purposes
        self.otherView = None

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:  # noqa: N802
        super().mousePressEvent(e)
        self.handle_expand(e)

    def mouseDoubleClickEvent(self, e: QtGui.QMouseEvent) -> None:  # noqa: N802
        super().mouseDoubleClickEvent(e)
        self.handle_expand(e)

    def handle_expand(self, e: QtGui.QMouseEvent) -> None:
        """
        Check if click happened on the part with cause expanding the item in the view. If so also expand/collapse
        the same item in the other view.
        """
        clicked_index = self.indexAt(e.pos())
        if clicked_index.isValid():
            # if click is left from text then it is expand event
            vrect = self.visualRect(clicked_index)
            item_identation = vrect.x() - self.visualRect(self.rootIndex()).x()
            if e.pos().x() < item_identation:
                self.expand_in_other_view(clicked_index, not self.isExpanded(clicked_index))

    def expand_in_other_view(self, clicked_index: QModelIndex, is_expanded: bool) -> None:
        """
        Expand/collapse item which is same to the clicked_index in the other view. The other view is one of the left
        if click happened on the right and vice versa.

        Parameters
        ----------
        clicked_index
            Index of the item that was expanded/collapsed
        is_expanded
            This flag is true if the item was expanded otherwise if false (collapsed).
        """
        source_model_this = self.model().sourceModel()
        source_model_other = self.otherView.model().sourceModel()
        clicked_model_index = self.model().mapToSource(clicked_index)

        # register expanding change
        source_model_this.set_expanded(clicked_model_index, is_expanded)

        clicked_name = source_model_this.node_from_index(clicked_model_index).name

        index_other = source_model_other.index_from_name(clicked_name, source_model_other.rootItem)
        if index_other is not None:
            # if same item (item with the same name) exists in the other view
            self.otherView.setExpanded(index_other, is_expanded)

    def setModel(self, model: QtCore.QAbstractItemModel) -> None:  # noqa: N802
        """
        Override setModel that we can connect expandItem signal. This signal is emitted when model
        expand the item.
        """
        super().setModel(model)
        self.model().sourceModel().expand_item.connect(self.setExpanded)

    def setExpanded(self, index: QModelIndex, expand: bool) -> None:  # noqa: N802
        """
        Set item expanded/collapsed.

        Parameters
        ----------
        index
            Item's index
        expand
            This flag is true if the item was expanded otherwise if false (collapsed).
        """
        self.model().sourceModel().set_expanded(index, expand)
        index = self.model().mapFromSource(index)
        if expand:
            self.expand(index)
        else:
            self.collapse(index)


class OWMarkerGenes(widget.OWWidget):
    name = "Marker Genes"
    icon = 'icons/OWMarkerGenes.svg'
    priority = 130

    replaces = ['orangecontrib.single_cell.widgets.owmarkergenes.OWMarkerGenes']

    class Warning(widget.OWWidget.Warning):
        using_local_files = widget.Msg("Can't connect to serverfiles. Using cached files.")

    class Outputs:
        genes = widget.Output("Genes", Table)

    want_main_area = True
    want_control_area = True

    auto_commit = Setting(True)
    selected_source = Setting("")
    selected_organism = Setting("")
    selected_root_attribute = Setting(0)

    settingsHandler = MarkerGroupContextHandler()  # noqa: N815
    selected_genes = settings.ContextSetting([])

    settings_version = 2

    _data = None
    _available_sources = None

    def __init__(self) -> None:
        super().__init__()

        # define the layout
        main_area = QWidget(self.mainArea)
        self.mainArea.layout().addWidget(main_area)
        layout = QGridLayout()
        main_area.setLayout(layout)
        layout.setContentsMargins(4, 4, 4, 4)

        # filter line edit
        self.filter_line_edit = QLineEdit()
        self.filter_line_edit.setPlaceholderText("Filter marker genes")
        layout.addWidget(self.filter_line_edit, 0, 0, 1, 3)

        # define available markers view
        self.available_markers_view = TreeView()
        box = gui.vBox(self.mainArea, "Available markers", addToLayout=False)
        box.layout().addWidget(self.available_markers_view)
        layout.addWidget(box, 1, 0, 2, 1)

        # create selected markers view
        self.selected_markers_view = TreeView()
        box = gui.vBox(self.mainArea, "Selected markers", addToLayout=False)
        box.layout().addWidget(self.selected_markers_view)
        layout.addWidget(box, 1, 2, 2, 1)

        self.available_markers_view.otherView = self.selected_markers_view
        self.selected_markers_view.otherView = self.available_markers_view

        # buttons
        box = gui.vBox(self.mainArea, addToLayout=False, margin=0)
        layout.addWidget(box, 1, 1, 1, 1)
        self.move_button = gui.button(box, self, ">", callback=self._move_selected)

        self._init_description_area(layout)
        self._init_control_area()
        self._load_data()

    def _init_description_area(self, layout: QLayout) -> None:
        """
        Function define an info area with description of the genes and add it to the layout.
        """
        box = gui.widgetBox(self.mainArea, "Description", addToLayout=False)
        self.descriptionlabel = QTextBrowser(
            openExternalLinks=True, textInteractionFlags=(Qt.TextSelectableByMouse | Qt.LinksAccessibleByMouse)
        )
        box.setMaximumHeight(self.descriptionlabel.fontMetrics().height() * (NUM_LINES_TEXT + 3))

        # description filed
        self.descriptionlabel.setText("Select a gene to see information.")
        self.descriptionlabel.setFrameStyle(QTextBrowser.NoFrame)
        # no (white) text background
        self.descriptionlabel.viewport().setAutoFillBackground(False)
        box.layout().addWidget(self.descriptionlabel)
        layout.addWidget(box, 3, 0, 1, 3)

    def _init_control_area(self) -> None:
        """
        Function defines dropdowns and the button in the control area.
        """
        box = gui.widgetBox(self.controlArea, 'Database', margin=0)
        self.source_index = -1
        self.db_source_cb = gui.comboBox(box, self, 'source_index')
        self.db_source_cb.activated[int].connect(self._set_db_source_index)

        box = gui.widgetBox(self.controlArea, 'Organism', margin=0)
        self.organism_index = -1
        self.group_cb = gui.comboBox(box, self, 'organism_index')
        self.group_cb.activated[int].connect(self._set_group_index)

        box = gui.widgetBox(self.controlArea, 'Group by', margin=0)
        self.group_by_cb = gui.comboBox(
            box, self, 'selected_root_attribute', items=GROUP_BY_ITEMS, callback=self._setup
        )

        gui.rubber(self.controlArea)
        gui.auto_commit(self.controlArea, self, "auto_commit", "Commit")

    def sizeHint(self):
        return super().sizeHint().expandedTo(QSize(900, 500))

    @property
    def available_sources(self) -> dict:
        return self._available_sources

    @available_sources.setter
    def available_sources(self, value: dict) -> None:
        """
        Set _available_sources variable, add them to dropdown, and select the source that was previously.
        """
        self._available_sources = value

        items = sorted(value.keys(), reverse=True)  # panglao first
        try:
            idx = items.index(self.selected_source)
        except ValueError:
            idx = -1

        self.db_source_cb.clear()
        self.db_source_cb.addItems(items)

        if idx != -1:
            self.source_index = idx
            self.selected_source = items[idx]
        elif items:
            self.source_index = min(max(self.source_index, 0), len(items) - 1)

        self._set_db_source_index(self.source_index)

    @property
    def data(self) -> Table:
        return self._data

    @data.setter
    def data(self, value: Table):
        """
        Set the source data. The data is then filtered on the first meta column (group).
        Select set dropdown with the groups and select the one that was selected previously.
        """
        self._data = value
        domain = value.domain

        if domain.metas:
            group = domain.metas[0]
            groupcol, _ = value.get_column_view(group)

            if group.is_string:
                group_values = list(set(groupcol))
            elif group.is_discrete:
                group_values = group.values
            else:
                raise TypeError("Invalid column type")
            group_values = sorted(group_values)  # human first

            try:
                idx = group_values.index(self.selected_organism)
            except ValueError:
                idx = -1

            self.group_cb.clear()
            self.group_cb.addItems(group_values)

            if idx != -1:
                self.organism_index = idx
                self.selected_organism = group_values[idx]
            elif group_values:
                self.organism_index = min(max(self.organism_index, 0), len(group_values) - 1)

            self._set_group_index(self.organism_index)

    def _load_data(self) -> None:
        """
        Collect available data sources (marker genes data sets).
        """
        self.Warning.using_local_files.clear()

        found_sources = {}
        try:
            found_sources.update(serverfiles.ServerFiles().allinfo(SERVER_FILES_DOMAIN))
        except requests.exceptions.ConnectionError:
            found_sources.update(serverfiles.allinfo(SERVER_FILES_DOMAIN))
            self.Warning.using_local_files()

        self.available_sources = {item.get('title').split(': ')[-1]: item for item in found_sources.values()}

    def _source_changed(self) -> None:
        """
        Respond on change of the source and download the data.
        """
        if self.available_sources:
            file_name = self.available_sources[self.selected_source]['filename']

            try:
                serverfiles.update(SERVER_FILES_DOMAIN, file_name)
            except requests.exceptions.ConnectionError:
                # try to update file. Ignore network errors.
                pass

            try:
                file_path = serverfiles.localpath_download(SERVER_FILES_DOMAIN, file_name)
            except requests.exceptions.ConnectionError as err:
                # Unexpected error.
                raise err
            self.data = Table.from_file(file_path)

    def _setup(self) -> None:
        """
        Setup the views with data.
        """
        self.closeContext()
        self.selected_genes = []
        self.openContext((self.selected_organism, self.selected_source))
        data_not_selected, data_selected = self._filter_data_group(self.data)

        # add model to available markers view
        group_by = GROUP_BY_ITEMS[self.selected_root_attribute]
        tree_model = TreeModel(data_not_selected, group_by)
        proxy_model = FilterProxyModel(self.filter_line_edit)
        proxy_model.setSourceModel(tree_model)

        self.available_markers_view.setModel(proxy_model)
        self.available_markers_view.selectionModel().selectionChanged.connect(
            partial(self._on_selection_changed, self.available_markers_view)
        )

        tree_model = TreeModel(data_selected, group_by)
        proxy_model = FilterProxyModel(self.filter_line_edit)
        proxy_model.setSourceModel(tree_model)
        self.selected_markers_view.setModel(proxy_model)

        self.selected_markers_view.selectionModel().selectionChanged.connect(
            partial(self._on_selection_changed, self.selected_markers_view)
        )
        self.selected_markers_view.model().sourceModel().data_added.connect(self._selected_markers_changed)
        self.selected_markers_view.model().sourceModel().data_removed.connect(self._selected_markers_changed)

        # update output and messages
        self._selected_markers_changed()

    def _filter_data_group(self, data: Table) -> Tuple[Table, Tuple]:
        """
        Function filter the table based on the selected group (Mouse, Human) and divide them in two groups based on
        selected_data variable.

        Parameters
        ----------
        data
            Table to be filtered

        Returns
        -------
        data_not_selected
            Data that will initially be in available markers view.
        data_selected
            Data that will initially be in selected markers view.
        """
        group = data.domain.metas[0]
        gvec = data.get_column_view(group)[0]

        if group.is_string:
            mask = gvec == self.selected_organism
        else:
            mask = gvec == self.organism_index
        data = data[mask]

        # divide data based on selected_genes variable (context)
        unique_gene_names = np.core.defchararray.add(
            data.get_column_view("Entrez ID")[0].astype(str), data.get_column_view("Cell Type")[0].astype(str)
        )
        mask = np.isin(unique_gene_names, self.selected_genes)
        data_not_selected = data[~mask]
        data_selected = data[mask]
        return data_not_selected, data_selected

    def commit(self) -> None:
        rows = self.selected_markers_view.model().sourceModel().rootItem.get_data_rows()
        if len(rows) > 0:
            metas = [r.metas for r in rows]
            data = Table.from_numpy(self.data.domain, np.empty((len(metas), 0)), metas=np.array(metas))
            # always false for marker genes data tables in single cell
            data.attributes[GENE_AS_ATTRIBUTE_NAME] = False
            # set taxonomy id in data.attributes
            data.attributes[TAX_ID] = MAP_GROUP_TO_TAX_ID.get(self.selected_organism, '')
            # set column id flag
            data.attributes[GENE_ID_COLUMN] = "Entrez ID"
            data.name = 'Marker Genes'
        else:
            data = None
        self.Outputs.genes.send(data)

    def _update_description(self, view: TreeView) -> None:
        """
        Upate the description about the gene. Only in case when one gene is selected.
        """
        selection = self._selected_rows(view)
        qmodel = view.model().sourceModel()

        if len(selection) > 1 or len(selection) == 0 or qmodel.node_from_index(selection[0]).data_row is None:
            self.descriptionlabel.setText("Select a gene to see information.")
        else:
            data_row = qmodel.node_from_index(selection[0]).data_row
            self.descriptionlabel.setHtml(
                f"<b>Gene name:</b> {data_row['Name']}<br/>"
                f"<b>Entrez ID:</b> {data_row['Entrez ID']}<br/>"
                f"<b>Cell Type:</b> {data_row['Cell Type']}<br/>"
                f"<b>Function:</b> {data_row['Function']}<br/>"
                f"<b>Reference:</b> <a href='{data_row['URL']}'>{data_row['Reference']}</a>"
            )

    # callback functions

    def _selected_markers_changed(self) -> None:
        """
        This function is called when markers in the selected view are added or removed.
        """
        rows = self.selected_markers_view.model().sourceModel().rootItem.get_data_rows()
        self.selected_genes = [row["Entrez ID"].value + row["Cell Type"].value for row in rows]
        self.commit()

    def _on_selection_changed(self, view: TreeView) -> None:
        """
        When selection in one of the view changes in a view button should change a sign in the correct direction and
        other view should reset the selection. Also gene description is updated.
        """
        self.move_button.setText(">" if view is self.available_markers_view else "<")
        if view is self.available_markers_view:
            self.selected_markers_view.clearSelection()
        else:
            self.available_markers_view.clearSelection()
        self._update_description(view)

    def _set_db_source_index(self, source_index: int) -> None:
        """
        Set the index of selected database source - index in a dropdown.
        """
        self.source_index = source_index
        self.selected_source = self.db_source_cb.itemText(source_index)
        self._source_changed()

    def _set_group_index(self, group_index: int) -> None:
        """
        Set the index of organism - index in a dropdown.
        """
        self.organism_index = group_index
        self.selected_organism = self.group_cb.itemText(group_index)
        self._setup()

    def _move_selected(self) -> None:
        """
        Move selected genes when button clicked.
        """
        if self._selected_rows(self.selected_markers_view):
            self._move_selected_from_to(self.selected_markers_view, self.available_markers_view)
        elif self._selected_rows(self.available_markers_view):
            self._move_selected_from_to(self.available_markers_view, self.selected_markers_view)

    # support functions for callbacks

    def _move_selected_from_to(self, src: TreeView, dst: TreeView) -> None:
        """
        Function moves items from src model to dst model.
        """
        selected_items = self._selected_rows(src)

        src_model = src.model().sourceModel()
        dst_model = dst.model().sourceModel()

        # move data as mimeData from source to destination tree view
        mime_data = src_model.mimeData(selected_items)
        # remove nodes from the source view
        src_model.remove_node_list(selected_items)
        dst_model.dropMimeData(mime_data, Qt.MoveAction, -1, -1)

    @staticmethod
    def _selected_rows(view: TreeView) -> List[QModelIndex]:
        """
        Return the selected rows in the view.
        """
        rows = view.selectionModel().selectedRows()
        return list(map(view.model().mapToSource, rows))

    @classmethod
    def migrate_settings(cls, settings, version=0):
        def migrate_to_version_2():
            settings["selected_source"] = settings.pop("selected_db_source", "")
            settings["selected_organism"] = settings.pop("selected_group", "")
            if "context_settings" in settings:
                for co in settings["context_settings"]:
                    co.values["selected_genes"] = [g[0] + g[1] for g in co.values["selected_genes"]]

        if version < 2:
            migrate_to_version_2()


if __name__ == "__main__":
    from orangewidget.utils.widgetpreview import WidgetPreview

    WidgetPreview(OWMarkerGenes).run()
